# coding: utf-8
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
Things that should run often.
"""

from gluish.utils import random_string
from siskin.task import DefaultTask
import datetime
import luigi

class Workflow(DefaultTask):
    TAG = 'workflows'

class DailyIndexUpdates(Workflow):
    """
    A wrapper around a lot of heavy tasks to keep ES up to date.
    """
    date = luigi.DateParameter(default=datetime.date.today())
    indicator = luigi.Parameter(default=random_string())

    def requires(self):
        from siskin.sources.bms import BMSIndex
        from siskin.sources.bnf import BNFIndex
        from siskin.sources.bsz import BSZIndex
        from siskin.sources.disson import DissonIndex
        from siskin.sources.doab import DOABIndex
        from siskin.sources.ebl import EBLIndex
        from siskin.sources.elsevier import ElsevierIndex
        from siskin.sources.ema import EMAIndex
        from siskin.sources.gbv import GBVIndex
        from siskin.sources.hszigr import HSZIGRIndex
        from siskin.sources.imslp import IMSLPIndex
        from siskin.sources.ksd import KSDIndex
        from siskin.sources.lfer import LFERIndex
        from siskin.sources.mor import MORIndex
        from siskin.sources.mtc import MTCIndex
        from siskin.sources.naxos import NaxosIndex
        from siskin.sources.nep import NEPIndex
        from siskin.sources.nl import NLIndex
        from siskin.sources.oso import OSOIndex
        from siskin.sources.pao import PAOIndex
        from siskin.sources.qucosa import QucosaIndex
        from siskin.sources.rism import RISMIndex

        return [
            BMSIndex(date=self.date),
            BNFIndex(date=self.date),
            BSZIndex(end=self.date, indicator=self.indicator),
            DissonIndex(date=self.date),
            DOABIndex(date=self.date),
            EBLIndex(date=self.date),
            ElsevierIndex(date=self.date),
            EMAIndex(date=self.date),
            GBVIndex(date=self.date),
            HSZIGRIndex(date=self.date),
            IMSLPIndex(date=self.date),
            KSDIndex(date=self.date),
            LFERIndex(date=self.date),
            MORIndex(date=self.date),
            MTCIndex(date=self.date),
            NaxosIndex(date=self.date),
            NEPIndex(date=self.date),
            NLIndex(date=self.date),
            OSOIndex(date=self.date),
            PAOIndex(date=self.date),
            QucosaIndex(date=self.date),
            RISMIndex(date=self.date),
        ]

    def run(self):
        self.logger.debug("Synced {0} indices on {1}".format(len(self.input()), self.date))
        with self.output().open('w'):
            pass

    def output(self):
        return luigi.LocalTarget(path=self.path())



# coding: utf-8

from gluish.intervals import yearly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.task import DefaultTask
import datetime
import json
import luigi

class BartocTask(DefaultTask):
    """ Bartoc. """
    TAG = 'bartoc'

    def closest(self):
        return yearly(self.date)

class BartocDump(BartocTask):
    """ Dump JSON. http://bartoc.org/en/node/761 """

    date = ClosestDateParameter(default=datetime.date.today())

    def run(self):
        output = shellout("""wget -O {output} 'http://bartoc.org/en/download/json?download=1&eid=767' """)
        with open(output) as handle:
            try:
                _ = json.load(handle)
            except ValueError as err:
                raise RuntimeError('invalid download: %s' % err)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='json'))



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,E1103
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
GND-related tasks.
"""

from elasticsearch import helpers as eshelpers
from gluish.common import Executable
from gluish.format import TSV
from gluish.intervals import monthly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from luigi.contrib.esindex import CopyToIndex
from siskin.benchmark import timed
from siskin.common import Directory
from siskin.database import sqlitedb
from siskin.sources.dbpedia import DBPGNDLinks, DBPDepictions
from siskin.task import DefaultTask
from siskin.utils import ElasticsearchMixin, random_string, iterfiles
import BeautifulSoup
import collections
import cPickle as pickle
import datetime
import elasticsearch
import hashlib
import HTMLParser
import itertools
import json
import luigi
import os
import prettytable
import re
import requests
import shutil
import string
import tempfile
import urllib

class GNDTask(DefaultTask):
    TAG = 'gnd'

    def closest(self):
        return monthly(self.date)

class DNBStatus(GNDTask):
    """ List available downloads. For human eyes only. """

    @timed
    def run(self):
        host = "datendienst.dnb.de"
        path = "/cgi-bin/mabit.pl"
        params = {
            "userID": "opendata",
            "pass": "opendata",
            "cmd": "login",
        }
        r = requests.get('http://{host}{path}?{params}'.format(host=host,
                         path=path, params=urllib.urlencode(params)))
        if not r.status_code == 200:
            raise RuntimeError(r)

        soup = BeautifulSoup.BeautifulSoup(r.text)
        tables = soup.findAll('table')
        if not tables:
            raise RuntimeError('No tabular data found on {0}'.format(r.url))

        table = tables[0]
        h = HTMLParser.HTMLParser()

        rows = []
        for tr in table.findAll('tr'):
            row = []
            for td in tr.findAll('td'):
                text = ''.join(h.unescape(td.find(text=True)))
                row.append(text)
            rows.append(row)

        x = prettytable.PrettyTable(rows[0])
        for row in rows[1:]:
            x.add_row(row)
        print(x)

    def complete(self):
        return False

class GNDDump(GNDTask):
    """ Download the GND snapshot. """

    date = ClosestDateParameter(default=datetime.date.today())
    format = luigi.Parameter(default='ttl', description='alternatively rdf')

    @timed
    def run(self):
        host = "datendienst.dnb.de"
        path = "/cgi-bin/mabit.pl"
        params = {
            "userID": "opendata",
            "pass": "opendata",
            "cmd": "fetch",
            "mabheft": "GND.%s.gz" % self.format,
        }

        url = 'http://{host}{path}?{params}'.format(host=host,
                                                    path=path,
                                                    params=urllib.urlencode(params))
        output = shellout("""wget --retry-connrefused "{url}" -O {output}""",
                          url=url)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ttl.gz'))

class GNDExtract(GNDTask):
    """ Extract the archive. """

    date = ClosestDateParameter(default=datetime.date.today())
    format = luigi.Parameter(default='ttl', description='alternatively rdf')

    def requires(self):
        return GNDDump(date=self.date, format=self.format)

    @timed
    def run(self):
        output = shellout("gunzip -c {input} > {output}", input=self.input().fn)
        luigi.File(output).move(self.output().fn)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ttl'))

class GNDCacheDB(GNDTask):
    """ Turn the dump into a (id, content) sqlite3 db.
    This artefact will be used by the cache server.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return GNDExtract(date=self.date, format='rdf')

    @timed
    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        pattern = re.compile("""rdf:about="http://d-nb.info/gnd/([0-9X-]+)">""")

        with sqlitedb(stopover) as cursor:
            cursor.execute("""CREATE TABLE gnd (id text  PRIMARY KEY, content blob)""")
            cursor.execute("""CREATE INDEX IF NOT EXISTS idx_gnd_id ON gnd (id)""")

            with self.input().open() as handle:
                groups = itertools.groupby(handle, key=str.isspace)
                for i, (k, lines) in enumerate(groups):
                    if k:
                        continue
                    lines = map(string.strip, list(lines))
                    match = pattern.search(lines[0])
                    if match:
                        row = (match.group(1), '\n'.join(lines))
                        cursor.execute("INSERT INTO gnd VALUES (?, ?)", row)

        luigi.File(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='db'))

class GNDNTriples(GNDTask):
    """ Get a Ntriples representation of GND. """

    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return GNDExtract(date=self.date)

    @timed
    def run(self):
        output = shellout("serdi -b -i turtle -o ntriples {input} > {output}",
                          input=self.input().path)
        # output = shellout("sort {input} > {output}", input=output)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='nt'))

class GNDTypes(GNDTask):
    """ Extract all rdf types. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return GNDNTriples(date=self.date)

    def run(self):
        output = shellout("""LANG=C grep -F 'http://www.w3.org/1999/02/22-rdf-syntax-ns#type' {input} > {output}""", input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class GNDFriends(GNDTask):
    """ Extract friendship relations. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return GNDNTriples(date=self.date)

    def run(self):
        output = shellout("""LANG=C grep -F 'http://d-nb.info/standards/elementset/gnd#acquaintanceshipOrFriendship' {input} > {output}""", input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class GNDFriendshipGraph(GNDTask):
    """ Create a simple DOT graph representation of the friendships. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return GNDFriends(date=self.date)

    def run(self):
        with self.input().open() as handle:
            with self.output().open('w') as output:
                output.write("graph friends {\n")
                for line in handle:
                    parts = line.split()
                    if not len(parts) == 4:
                        continue
                    s, _, p, _ = parts
                    s = s.replace("http://d-nb.info/gnd/", "").strip("<>")
                    p = p.replace("http://d-nb.info/gnd/", "").strip("<>")
                    output.write('\t"%s" -- "%s";\n' % (s, p))
                output.write("}\n")

    def output(self):
        return luigi.LocalTarget(path=self.path())

class GNDFriendshipImage(GNDTask):
    """ Create an image of the graph. """
    date = ClosestDateParameter(default=datetime.date.today())
    format = luigi.Parameter(default='svg')

    def requires(self):
        return GNDFriendshipGraph(date=self.date)

    @timed
    def run(self):
        output = shellout("dot -T{format} -o {output} {input}", format=self.format, input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext=self.format))

class GNDDates(GNDTask):
    """ Extract all dateOf... links. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return GNDNTriples(date=self.date)

    def run(self):
        output = shellout("""LANG=C grep -F 'http://d-nb.info/standards/elementset/gnd#dateOf' {input} > {output}""", input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class GNDBiographicalData(GNDTask):
    """ For each GND collect birth date and death dates. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return GNDDates(date=self.date)

    def run(self):
        biomap = collections.defaultdict(dict)
        with self.input().open() as handle:
            for line in handle:
                line = line.strip()
                parts = line.split()
                if len(parts) < 4:
                    raise RuntimeError('invalid ntriple: %s' % line)
                s, p, o = parts[0], parts[1], ' '.join(parts[2:-1])
                gnd = s.split('/')[-1].strip('<>')
                value = 'NOT_AVAILABLE'
                if o.endswith('^^<http://www.w3.org/2001/XMLSchema#date>'):
                    try:
                        value = datetime.date(*map(int, o[1:11].split('-')))
                    except ValueError as err:
                        self.logger.debug(line)
                        self.logger.warn(err)
                        continue

                elif o.endswith('^^<http://www.w3.org/2001/XMLSchema#gYear>'):
                    try:
                        value = datetime.date(int(o[1:5]), 1, 1)
                    except ValueError as err:
                        self.logger.debug(line)
                        self.logger.warn(err)
                        continue

                elif len(o.strip('"')) == 4:
                    if o.strip('"').isdigit():
                        value = datetime.date(int(o.strip('"')), 1, 1)
                else:
                    pass
                    # self.logger.debug('unknown type: %s' % line)

                if p == '<http://d-nb.info/standards/elementset/gnd#dateOfBirth>':
                    biomap[gnd]['b'] = value
                if p == '<http://d-nb.info/standards/elementset/gnd#dateOfDeath>':
                    biomap[gnd]['d'] = value

        with self.output().open('w') as output:
            for gnd, bio in biomap.iteritems():
                output.write_tsv(gnd, bio.get('b', 'NOT_AVAILABLE'), bio.get('d', 'NOT_AVAILABLE'))

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class GNDContemporaryPeople(GNDTask):
    """ Build equivalence classes of people with birthdates within a given timespan. """
    date = ClosestDateParameter(default=datetime.date.today())
    threshold = luigi.FloatParameter(default=0.7)

    def requires(self):
        return GNDBiographicalData(date=self.date)

    @timed
    def run(self):
        yearspan = collections.defaultdict(set)
        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('gnd', 'birth', 'death')):
                try:
                    birth = datetime.date(*map(int, row.birth.split('-')))
                    death = datetime.date(*map(int, row.death.split('-')))
                except ValueError:
                    continue

                yearspan[row.gnd] = set(range(birth.year, death.year + 1))

        with self.output().open('w') as output:
            for gnd, span in yearspan.iteritems():
                for other, o in yearspan.iteritems():
                    union = span.union(o)
                    if len(union) == 0:
                        continue
                    jaccard = len(span.intersection(o)) / float(len(union))
                    if jaccard > self.threshold:
                        output.write_tsv(gnd, other, jaccard)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class GNDFoafPages(GNDTask):
    """ Extract all FOAF links. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return GNDNTriples(date=self.date)

    def run(self):
        output = shellout("""LANG=C grep -F 'http://xmlns.com/foaf/0.1/page' {input} > {output}""", input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class GNDFoafLinks(GNDTask):
    """ Rewrite the GNDFoafPages output, so that we can compare it to
    the output of `DBPGNDValidLinks`. """

    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return GNDFoafPages(date=self.date)

    def run(self):
        output = shellout("""cut -d ' ' -f1,3 {input} |
                             sed -e 's@http://d-nb.info/gnd/@gnd:@g; s@http://de.wikipedia.org/wiki/@dbp:@g;' |
                             tr -d '<>' | awk '{{print $2" "$1}}' | tr ' ' '\\t' > {output}""", input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class GNDFoafOnly(GNDTask):
    """ Which links to dbpedia in GND are only via foaf:page? """

    date = ClosestDateParameter(default=datetime.date.today())
    version = luigi.Parameter(default="2014")
    language = luigi.Parameter(default="de")

    def requires(self):
        from siskin.sources.dbpedia import DBPGNDValidLinks
        return {'sa': DBPGNDValidLinks(version=self.version, language=self.language),
                'foaf': GNDFoafLinks(date=self.date)}

    def run(self):
        sa = set()
        with self.input().get('sa').open() as handle:
            for row in handle.iter_tsv(cols=('dbp', 'gnd')):
                sa.add((row.dbp, row.gnd))

        foaf = set()
        with self.input().get('foaf').open() as handle:
            for row in handle.iter_tsv(cols=('dbp', 'gnd')):
                foaf.add((row.dbp, row.gnd))

        with self.output().open('w') as output:
            for dbp, gnd in foaf.difference(sa):
                output.write_tsv(dbp, gnd)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class GNDGeonames(GNDTask):
    """ Extract all geonames from dump """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return GNDNTriples(date=self.date)

    def run(self):
        output = shellout("""LANG=C grep -F 'http://sws.geonames.org/' {input} |
                             cut -d ' ' -f3 | tr -d '<>' | sort -u > {output}""", input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class GNDList(GNDTask):
    """ Just dump a list of uniq GNDs. No URIs, just one id per line. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return GNDNTriples(date=self.date)

    @timed
    def run(self):
        output = shellout("""LANG=C awk '{{print $1}}' {input} |
                             LANG=C sed -e 's@<http://d-nb.info/gnd/@@g' |
                             LANG=C sed -e 's@>@@g' |
                             LANG=C grep -v '^_:' | sort -u > {output}""", input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class GNDNames(GNDTask):
    """ Extract all preferred and variant names and list them as (gnd, name) tuples. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return GNDNTriples(date=self.date)

    @timed
    def run(self):
        output = shellout("""LANG=C grep -E "variantName|preferredName" {input} |
                             LANG=C sed -e 's@<http://d-nb.info/gnd/\([0-9X-]*\)>@\\1@g' |
                             LANG=C sed -e 's@<http://d-nb.info/standards/elementset/gnd#[^>]*>@@g' |
                             LANG=C sed 's/.$//' | sed -e 's/"//g' | LANG=C grep -v "_:node" > {output}""", input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class GNDNamesWikipediaTitles(GNDTask):
    """ Match GND names with wikipedia page titles to find more connections. """
    date = ClosestDateParameter(default=datetime.date.today())
    language = luigi.Parameter(default='de')

    def requires(self):
        from siskin.sources.wikipedia import WikipediaTitles
        return {'gnd': GNDNames(date=self.date),
                'titles': WikipediaTitles(language=self.language)}

    @timed
    def run(self):
        titles = set()
        with self.input().get('titles').open() as handle:
            for row in handle.iter_tsv(cols=('name',)):
                titles.add(row.name)

        _, stopover = tempfile.mkstemp(prefix='siskin-')
        with self.input().get('gnd').open() as handle:
            with luigi.File(stopover, format=TSV).open('w') as output:
                for line in handle:
                    parts = line.split()
                    if len(parts) == 0:
                        continue
                    gnd, name = parts[0], string.strip(' '.join(parts[1:]))
                    output.write_tsv('HIT' if name in titles else 'MISS', gnd, name)

        output = shellout("sort -u {input} > {output}", input=stopover)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class GNDCount(GNDTask):
    """ Just count the number of triples and store it. """

    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return GNDNTriples(date=self.date)

    @timed
    def run(self):
        output = shellout("wc -l {input} | awk '{{print $1}}' > {output}", input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='count'))

class GNDNTriplesSplitted(GNDTask):
    """ Split Ntriples into chunks, so we see some progress in virtuoso. """

    date = ClosestDateParameter(default=datetime.date.today())
    lines = luigi.IntParameter(default=1000000)

    def requires(self):
        return GNDNTriples(date=self.date)

    @timed
    def run(self):
        prefix = '{0}-'.format(random_string())
        output = shellout("cd {tmp} && split -l {lines} -a 8 {input} {prefix} && cd -",
                          lines=self.lines, tmp=tempfile.gettempdir(),
                          input=self.input().path, prefix=prefix)
        target = os.path.join(self.taskdir())
        if not os.path.exists(target):
            os.makedirs(target)
        with self.output().open('w') as output:
            for path in iterfiles(tempfile.gettempdir()):
                filename = os.path.basename(path)
                if filename.startswith(prefix):
                    dst = os.path.join(target, filename)
                    shutil.move(path, dst)
                    output.write_tsv(dst)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='filelist'), format=TSV)

class GNDVirtuoso(GNDTask):
    """ Load GND parts into virtuoso 6.
    Clear any graph manually with `SPARQL CLEAR GRAPH <http://dbpedia.org/resource/>;`

    See SELECT * FROM DB.DBA.LOAD_LIST; for progress.

    Also add `dirname $(taskoutput GNDNTriplesSplitted)` to
    AllowedDirs in /etc/virtuoso-opensource-6.1/virtuoso.ini
    """

    date = ClosestDateParameter(default=datetime.date.today())
    graph = luigi.Parameter(default='http://d-nb.info/gnd/')
    host = luigi.Parameter(default='localhost', significant=False)
    port = luigi.IntParameter(default=1111, significant=False)
    username = luigi.Parameter(default='dba', significant=False)
    password = luigi.Parameter(default='dba', significant=False)

    def requires(self):
        return GNDNTriplesSplitted(date=self.date)

    @timed
    def run(self):
        with self.input().open() as handle:
            path = handle.iter_tsv(cols=('path',)).next().path
            dirname = os.path.dirname(path)
            filename = os.path.basename(path)
            prefix, _ = filename.split('-')

        print("""
            Note: If you want to reload everything, clear any previous graph first:

                log_enable(3,1);
                SPARQL CLEAR GRAPH <{graph}>;

            -----------------------------------------------------

            This is a manual task for now. If you haven't already:

            Run `isql-vt` (V6) or `isql` (V7), then execute:

                LD_DIR('{dirname}', '{prefix}-*', '{graph}');

            Check the result via:

                SELECT * FROM DB.DBA.LOAD_LIST;

            When you are done adding all triples files to the load list, run:

                RDF_LOADER_RUN();

            See also:

            * http://docs.openlinksw.com/virtuoso/fn_log_enable.html
            * http://docs.openlinksw.com/virtuoso/sparqlextensions.html#rdfsparulexamples10

            Note that "Section 16.3.2.3.10. Example for Dropping graph"
            calls graphs with 600M triples very large RDF data collections.

        """.format(dirname=dirname, prefix=prefix, graph=self.graph))

    def complete(self):
        return False

class GNDAbbreviatedNTriples(GNDTask):
    """ Get a Ntriples representation of GND, but abbreviate with ntto. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return GNDNTriples(date=self.date)

    @timed
    def run(self):
        output = shellout("ntto -a -r {rules} -o {output} {input}", input=self.input().path, rules=self.assets('RULES.txt'))
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='nt'))

class GNDPredicateDistribution(GNDTask):
    """ Just a uniq -c on the predicate 'column' """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return GNDNTriples(date=self.date)

    def run(self):
        output = shellout("""cut -d " " -f2 {input} | LANG=C sort | LANG=C uniq -c > {output}""",
                          input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='txt'))

class GNDCayleyLevelDB(GNDTask):
    """ Create a Cayley LevelDB database from GND data. """
    date = ClosestDateParameter(default=datetime.date.today())
    gomaxprocs = luigi.IntParameter(default=8, significant=False)

    def requires(self):
        return {'ntriples': GNDAbbreviatedNTriples(date=self.date),
                'cayley': Executable(name='cayley', message='http://git.io/KH-wFA')}

    @timed
    def run(self):
        dbpath = tempfile.mkdtemp(prefix='siskin-')
        shellout("cayley init -alsologtostderr -config {config} -dbpath={dbpath}",
                 config=self.assets('cayley.leveldb.conf'), dbpath=dbpath)
        shellout("GOMAXPROCS={gomaxprocs} cayley load -config {config} -alsologtostderr -dbpath={dbpath} --triples {input}",
                 gomaxprocs=self.gomaxprocs, config=self.assets('cayley.leveldb.conf'), dbpath=dbpath, input=self.input().get('ntriples').path)
        shutil.move(dbpath, self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='leveldb'))

class GNDCayleyBoltDB(GNDTask):
    """ Create a Cayley BoltDB database from GND data. """
    date = ClosestDateParameter(default=datetime.date.today())
    gomaxprocs = luigi.IntParameter(default=8, significant=False)

    def requires(self):
        return {'ntriples': GNDAbbreviatedNTriples(date=self.date),
                'cayley': Executable(name='cayley', message='http://git.io/KH-wFA')}

    @timed
    def run(self):
        _, dbpath = tempfile.mkstemp(prefix='siskin-')
        shellout("cayley init -alsologtostderr -config {config} -dbpath={dbpath}",
                 config=self.assets('cayley.bolt.conf'), dbpath=dbpath)
        shellout("GOMAXPROCS={gomaxprocs} cayley load -config {config} -alsologtostderr -dbpath={dbpath} --triples {input}",
                 gomaxprocs=self.gomaxprocs, config=self.assets('cayley.bolt.conf'), dbpath=dbpath, input=self.input().get('ntriples').path)
        shutil.move(dbpath, self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='bolt'))

class GNDCayleyMongoDB(GNDTask):
    """ Create a Cayley MongoDB database from GND and dbpedia data. """
    date = ClosestDateParameter(default=datetime.date.today())
    gomaxprocs = luigi.IntParameter(default=8, significant=False)

    def requires(self):
        return {'ntriples': GNDAbbreviatedNTriples(date=self.date),
                'cayley': Executable(name='cayley', message='http://git.io/KH-wFA')}

    @timed
    def run(self):
        shellout("cayley init -alsologtostderr -config {config}",
                 config=self.assets('cayley.mongodb.conf'))
        shellout("cayley load -config {config} -alsologtostderr --triples {input}",
                 config=self.assets('cayley.mongodb.conf'),
                 input=self.input().get('ntriples').path)
        shellout("touch {output}", output=self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='mongo'))

class GNDJson(GNDTask):
    """ Convert GND to some indexable JSON. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return GNDNTriples(date=self.date)

    @timed
    def run(self):
        output = shellout("ntto -r {rules} -a -j -i {input} > {output}", input=self.input().path, rules=self.assets('RULES.txt'))
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'))

class GNDIndex(GNDTask, CopyToIndex):
    date = ClosestDateParameter(default=datetime.date.today())

    index = 'gnd'
    doc_type = 'triples'
    purge_existing_index = True

    def requires(self):
        return GNDJson(date=self.date)

class GNDTaxonomy(GNDTask):
    """
    For each term in broaderTermGeneral, save broader and narrower terms.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return GNDRelations(date=self.date, relation='dnb:broaderTermGeneral')

    def run(self):
        taxonomy = {}
        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('narrower', 'broader')):
                taxonomy[row.narrower] = row.broader
        with self.output().open('w') as output:
            for key, value in taxonomy.iteritems():
                broader = []
                while True:
                    broader.append(value)
                    if value not in taxonomy:
                        break
                    else:
                        value = taxonomy[value]
                        if value in broader:
                            break
                output.write_tsv(key, "|".join(broader))

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class GNDDBPediaLinks(GNDTask):
    """ Return (s, o) tuples that connect gnd and wikipedia. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return GNDRelations(date=self.date, relation='owl:sameAs')

    def run(self):
        output = shellout(r""" awk '$2 ~ "dbp:" {{print $0}}' {input} > {output}""", input=self.input().path)
        with luigi.File(output, format=TSV).open() as handle:
            with self.output().open('w') as output:
                for row in handle.iter_tsv(cols=('gnd', 'dbp')):
                    decoded = urllib.unquote(row.dbp)
                    output.write_tsv(decoded, row.gnd)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class GNDDBPediaInterlinks(GNDTask):
    """
    Report GND <-> dbpedia interlinking stats.
    """

    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {
            'gndto': GNDDBPediaLinks(date=self.date),
            'dbpto': DBPGNDLinks(),
            'gndpage': GNDPage(date=self.date, relation='foaf:Page')}

    def run(self):
        gndto = set()
        with self.input().get('gndto').open() as handle:
            for row in handle.iter_tsv(cols=('gnd', 'dbp')):
                gndto.add(row.gnd)

        dbpto = set()
        with self.input().get('dbpto').open() as handle:
            for row in handle.iter_tsv(cols=('gnd', 'p', 'dbp')):
                dbpto.add(row.gnd)

        gndpage = set()
        with self.input().get('gndpage').open() as handle:
            for row in handle.iter_tsv(cols=('gnd', 'url')):
                dbpto.add(row.gnd)

        result = dict(
            gnd_links=len(gndto),
            dbp_links=len(dbpto),
            gnd_and_dbp=len(gndto.intersection(dbpto)),
            gnd_only=len(gndto - dbpto),
            dbp_only=len(dbpto - gndto),
            gnd_or_dbp=len(gndto.union(dbpto)),
            dbp_and_via_url=len(dbpto.intersection(gndpage)),
        )
        with self.output().open('w') as output:
            for key, value in result.iteritems():
                output.write_tsv(key, value)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class GNDDepictions(GNDTask):
    """
    For each GND, that has a DBP link (TODO: more is possible via backlinks)
    report the image URL in foaf:depiction.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {'gndto': GNDDBPediaLinks(date=self.date),
                'dbppics': DBPDepictions()}

    @timed
    def run(self):
        kv = shellout(""" tabtokv -f "1,3" -o {output} {input}""", input=self.input().get('dbppics').path)
        with self.input().get('gndto').open() as handle:
            with sqlitedb(kv) as cursor:
                with self.output().open('w') as output:
                    for row in handle.iter_tsv(cols=('dbp', 'gnd')):
                        cursor.execute("""select value from store where key = ?""", (row.dbp,))
                        result = cursor.fetchall()
                        for url in set(result):
                            output.write_tsv(row.gnd, url[0])

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class GNDDownloadDepictions(GNDTask):
    """ Download depictions from wikimedia, that are linked with a GND, about 70k. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {'urls': GNDDepictions(date=self.date), 'dir': Directory(path=self.taskdir())}

    def run(self):
        with self.input().get('urls').open() as handle:
            for i, row in enumerate(handle.iter_tsv(cols=('gnd', 'url'))):
                gnd = row.gnd.replace('gnd:', '')
                _, ext = os.path.splitext(row.url)
                imgname = '%s-%s%s' % (gnd, hashlib.sha1(row.url).hexdigest(), ext)
                imgpath = os.path.join(self.taskdir(), imgname)
                if not os.path.exists(imgpath):
                    output = shellout("""wget -q -O {output} "{url}" """, output=imgpath, url=row.url, ignoremap={8: '404s throw 8'})

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class GNDDepictionsDB(GNDTask):
    """ Create a small KV store. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return GNDDepictions(date=self.date)

    def run(self):
        output = shellout(""" tabtokv -f "1,2" -o {output} {input} """, input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='db'))

class GNDPredicates(GNDTask):
    """ Report all existing predicates. """
    date = ClosestDateParameter(default=datetime.date.today())
    index = luigi.Parameter(default='gnd', description='name of the index to search')

    def requires(self):
        return Executable(name='estab', message='http://git.io/bLY7cQ')

    @timed
    def run(self):
        output = shellout(r""" estab -indices {index} -f "p" | sort -u > {output} """, index=self.index)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(digest=True), format=TSV)

class GNDDatabase(GNDTask):
    """
    Build a relational version of GND. sqlite3, each relation
    gets an indexed (s, o) table.
    """

    date = ClosestDateParameter(default=datetime.date.today())
    index = luigi.Parameter(default='gnd', description='name of the index to search')

    def requires(self):
        relations = [
            "dnb:acquaintanceshipOrFriendship",
            "dnb:affiliation",
            "dnb:annotator",
            "dnb:architect",
            "dnb:broaderTermGeneral",
            "dnb:definition",
            "dnb:familialRelationship",
            "dnb:fieldOfActivity",
            "dnb:fieldOfStudy",
            "dnb:functionOrRole",
            "dnb:memberOfTheFamily",
            "dnb:placeOfActivity",
            "dnb:placeOfBirth",
            "dnb:placeOfExile",
            "dnb:playedInstrument",
            # "dnb:preferredNameForThePerson",
            "dnb:professionOrOccupation",
            "dnb:professionalRelationship",
            "dnb:relatedTerm",
            "dnb:related",
            # "foaf:Page",
            # "skos:narrowMatch",
        ]
        tasks = {}
        for relation in relations:
            tasks[relation] = GNDRelations(date=self.date, index=self.index, relation=relation)
        return tasks

    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        with sqlitedb(stopover) as cursor:
            for relation, target in self.input().iteritems():
                table = relation.replace(':', '_')
                cursor.execute("""CREATE TABLE IF NOT EXISTS %s (s TEXT, o TEXT)""" % table)
                with target.open() as handle:
                    for row in handle.iter_tsv(cols=('s', 'o')):
                        cursor.execute("""INSERT INTO %s (s, o) VALUES (?, ?)""" % table, row)
                cursor.connection.commit()
                cursor.execute("""CREATE INDEX IF NOT EXISTS idx_%s_s on %s (s)""" % (table, table))
                cursor.execute("""CREATE INDEX IF NOT EXISTS idx_%s_o on %s (o)""" % (table, table))
                cursor.execute("""CREATE INDEX IF NOT EXISTS idx_%s_s_o on %s (s, o)""" % (table, table))
                cursor.execute("""CREATE INDEX IF NOT EXISTS idx_%s_o_s on %s (o, s)""" % (table, table))
                cursor.connection.commit()
        luigi.File(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='db'))

class GNDRelations(GNDTask):
    """
    Extract relations from GND.
    Given the predicate shortname, extract (s, o) tuples from the index.

    Example relations:

    * dnb:acquaintanceshipOrFriendship
    * dnb:affiliation
    * dnb:annotator
    * dnb:architect
    * dnb:broaderTermGeneral
    * dnb:definition
    * dnb:familialRelationship
    * dnb:fieldOfActivity
    * dnb:fieldOfStudy
    * dnb:functionOrRole
    * dnb:memberOfTheFamily
    * dnb:placeOfActivity
    * dnb:placeOfExile
    * dnb:playedInstrument
    * dnb:preferredNameForThePerson
    * dnb:relatedTerm
    * foaf:Page
    * skos:narrowMatch
    """

    date = ClosestDateParameter(default=datetime.date.today())
    relation = luigi.Parameter(default='dnb:definition')
    index = luigi.Parameter(default='gnd', description='name of the index to search')

    def requires(self):
        return Executable(name='estab', message='http://git.io/bLY7cQ')

    def run(self):
        output = shellout(r""" estab -indices {index} -f "s o"
            -query '{{"query": {{"query_string": {{"query": "p:\"{relation}\""}}}}}}' > {output}""", index=self.index, relation=self.relation)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(digest=True), format=TSV)

class GNDRelationsDB(GNDTask):
    """ Build an sqlite key-value database from a single relation in the GND. """
    date = ClosestDateParameter(default=datetime.date.today())
    relation = luigi.Parameter(default='dnb:definition')
    index = luigi.Parameter(default='gnd', description='name of the index to search')

    def requires(self):
        return GNDRelations(date=self.date, relation=self.relation, index=self.index)

    @timed
    def run(self):
        output = shellout(r""" tabtokv -f "1,2" -o {output} {input} """, input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(digest=True))



# coding: utf-8
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#
#
# Aggregated Index workflows
#
# Configuration and filter notes
#
# Filters are sets of records and these sets should be combinable.
# Imaginary configuration structure:
#
# {
#     "DE-15": {
#         "filters": {
#             "and": {
#                 "or": {
#                     "holdings": {
#                         "file": "/path/to/h.xml"
#                     },
#                     "source": {
#                         "id": "333"
#                     },
#                 }
#                 "collections": {
#                     "file": "/path/to/collection/list"
#                 },
#             }
#         }
#     }
# }
#

from siskin.benchmark import timed
from gluish.common import Executable
from gluish.format import TSV
from gluish.intervals import weekly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.sources.crossref import CrossrefIntermediateSchema, CrossrefUniqISSNList
from siskin.sources.degruyter import DegruyterIntermediateSchema, DegruyterISSNList
from siskin.sources.doaj import DOAJIntermediateSchema, DOAJISSNList
from siskin.sources.doi import DOIBlacklist
# from siskin.sources.gbi import GBIIntermediateSchema, GBIISSNList
from siskin.sources.holdings import HoldingsFile
from siskin.sources.jstor import JstorIntermediateSchema, JstorISSNList
from siskin.task import DefaultTask
import datetime
import itertools
import luigi
import tempfile

class AITask(DefaultTask):
    TAG = 'ai'

    def closest(self):
        return weekly(self.date)

class DownloadFile(AITask):
    """ Download a file. TODO(miku): move this out here. """

    date = luigi.DateParameter(default=datetime.date.today())
    url = luigi.Parameter()

    @timed
    def run(self):
        output = shellout("""curl -L --fail "{url}" > {output}""", url=self.url)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='file', digest=True))

class DownloadAndUnzipFile(AITask):
    """ Download a file and unzip it. TODO(miku): move this out here. """
    date = luigi.DateParameter(default=datetime.date.today())
    url = luigi.Parameter()

    @timed
    def run(self):
        output = shellout("""curl -L --fail "{url}" > {output}""", url=self.url)
        output = shellout("""unzip -qq -p {input} > {output}""", input=output)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='file', digest=True))

class AIISSNStats(AITask):
    """ Match ISSN lists. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {
            'crossref': CrossrefUniqISSNList(date=self.date),
            'degruyter': DegruyterISSNList(date=self.date),
            'doaj': DOAJISSNList(date=self.date),
            # 'gbi': GBIISSNList(date=self.date),
            'jstor': JstorISSNList(date=self.date)
        }

    @timed
    def run(self):
        def loadset(target):
            s = set()
            with target.open() as handle:
                for row in handle.iter_tsv(cols=('issn',)):
                    s.add(row.issn)
            return s

        with self.output().open('w') as output:
            for k1, k2 in itertools.combinations(self.input().keys(), 2):
                s1 = loadset(self.input().get(k1))
                s2 = loadset(self.input().get(k2))
                output.write_tsv(k1, k2, len(s1), len(s2), len(s1.intersection(s2)))

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class AIISSNOverlaps(AITask):
    """ Match ISSN lists. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {
            'crossref': CrossrefUniqISSNList(date=self.date),
            'degruyter': DegruyterISSNList(date=self.date),
            'doaj': DOAJISSNList(date=self.date),
            # 'gbi': GBIISSNList(date=self.date),
            'jstor': JstorISSNList(date=self.date)
        }

    @timed
    def run(self):
        def loadset(target):
            s = set()
            with target.open() as handle:
                for row in handle.iter_tsv(cols=('issn',)):
                    s.add(row.issn)
            return s

        with self.output().open('w') as output:
            for k1, k2 in itertools.combinations(self.input().keys(), 2):
                s1 = loadset(self.input().get(k1))
                s2 = loadset(self.input().get(k2))
                for issn in sorted(s1.intersection(s2)):
                    output.write_tsv(k1, k2, issn)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class AIIntermediateSchema(AITask):
    """ Create an intermediate schema record from all AI sources. """

    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return [Executable(name='pigz', message='http://zlib.net/pigz/'),
                CrossrefIntermediateSchema(date=self.date),
                DegruyterIntermediateSchema(date=self.date),
                DOAJIntermediateSchema(date=self.date),
                GBIIntermediateSchema(date=self.date),
                JstorIntermediateSchema(date=self.date)]

    @timed
    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        for target in self.input()[1:]:
            shellout("cat {input} | pigz >> {output}", input=target.path, output=stopover)
        luigi.File(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'))

class AIExport(AITask):
    """ Create a SOLR-importable file. """

    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {
            'crossref': CrossrefIntermediateSchema(date=self.date),
            'degruyter': DegruyterIntermediateSchema(date=self.date),
            'doaj': DOAJIntermediateSchema(date=self.date),
            'jstor' : JstorIntermediateSchema(date=self.date),

            # 'gbi': GBIIntermediateSchema(date=self.date),

            'DE-105': DownloadAndUnzipFile(date=self.date, url='https://goo.gl/Gq199T'),
            'DE-14': DownloadAndUnzipFile(date=self.date, url='https://goo.gl/Tz3vbk'),
            'DE-15': DownloadAndUnzipFile(date=self.date, url='https://goo.gl/inyKLr'),
            'DE-Bn3': DownloadAndUnzipFile(date=self.date, url='https://goo.gl/oq8LDD'),
            'DE-Ch1': DownloadAndUnzipFile(date=self.date, url='https://goo.gl/uJwoUf'),
            'DE-Gla1': DownloadAndUnzipFile(date=self.date, url='https://goo.gl/6506Dz'),
            'DE-Zi4': DownloadAndUnzipFile(date=self.date, url='https://goo.gl/Ld0LCw'),
            'DE-J59': DownloadAndUnzipFile(date=self.date, url='https://goo.gl/44xEbF'),

            'DE-15-FID': DownloadFile(date=self.date, url='https://goo.gl/tm6U9D'),

            'blacklist': DOIBlacklist(date=self.date),
            'app': Executable(name='span-export', message='http://git.io/vI8NV'),
            'pigz': Executable(name='pigz', message='http://zlib.net/pigz/'),
        }

    @timed
    def run(self):
        """
        TODO(miku): Move source/ISIL matrix and filter method out of here.
        """
        _, stopover = tempfile.mkstemp(prefix='siskin-')

        # DE-15 and DE-14 get DOAJ, cf. #6524, #4983
        shellout("""span-export -doi-blacklist {blacklist} -any DE-540 -l DE-15-FID:{issns} -any DE-15 -any DE-14 {input} | pigz >> {output}""",
                 blacklist=self.input().get('blacklist').path, input=self.input().get('doaj').path, issns=self.input().get('DE-15-FID').path, output=stopover)

        # DE-15 gets DeGruyter, cf. #4731
        shellout("""span-export -doi-blacklist {blacklist} -skip -f DE-15:{holding} {input} | pigz >> {output}""",
                 blacklist=self.input().get('blacklist').path, holding=self.input().get('DE-15').path, input=self.input().get('degruyter').path, output=stopover)

        # JSTOR detached from all for now, cf. #5472
        shellout("""span-export -doi-blacklist {blacklist} {input} | pigz >> {output}""",
                 blacklist=self.input().get('blacklist').path, input=self.input().get('jstor').path, output=stopover)

        def format_args(flagname, isils):
            """ Helper to create args like -f DE-15:/path/to/target -f DE-14:/path/to/target ..."""
            args = ["%s %s:%s" % (flagname, isil, self.input().get(isil).path) for isil in isils]
            return " ".join(args)

        files = format_args("-f", ['DE-105', 'DE-14', 'DE-15', 'DE-Bn3', 'DE-Ch1', 'DE-Gla1', 'DE-Zi4', 'DE-J59'])
        lists = format_args("-l", ['DE-15-FID'])

        # apply holdings and issn filters on sources
        for source in ('crossref'):
            shellout("span-export -doi-blacklist {blacklist} -skip {files} {lists} {input} | pigz >> {output}",
                     blacklist=self.input().get('blacklist').path, files=files, lists=lists,
                     input=self.input().get(source).path, output=stopover)

        luigi.File(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'))

class AIUpdate(AITask, luigi.WrapperTask):
    """
    Just a wrapper task for updates, refs #5702.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return [AIExport(date=self.date), AIIntermediateSchema(date=self.date)]

    def output(self):
        return self.input()



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,E1103,C0301
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
Freebase.
"""

from siskin.benchmark import timed
from gluish.common import Executable
from gluish.intervals import monthly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.task import DefaultTask
import datetime
import luigi
import tempfile

class FreebaseTask(DefaultTask):
    TAG = 'freebase'

    def closest(self):
        return monthly(date=self.date)

class FreebasePublic(FreebaseTask):
    """ Download bucket descriptor. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return Executable(name='wget')

    @timed
    def run(self):
        url = "http://commondatastorage.googleapis.com/freebase-public/"
        output = shellout(""" wget --retry-connrefused {url} -O {output}""", url=url)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='xml'))

class FreebaseDump(FreebaseTask):
    """ Download the dump. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return Executable(name='wget')

    @timed
    def run(self):
        url = "http://commondatastorage.googleapis.com/freebase-public/rdf/freebase-rdf-latest.gz"
        output = shellout(""" wget -q --retry-connrefused {url} -O {output}""", url=url)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='gz'))

class FreebaseExtract(FreebaseTask):
    """ Extract the dump. 250G+ !"""
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return FreebaseDump(date=self.date)

    @timed
    def run(self):
        output = shellout(""" gunzip -c {input} > {output}""",
                          input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='nt'))

class FreebaseAbbreviatedNTriples(FreebaseTask):
    """ Convert all Freebase ntriples to abbreviated ones. """

    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return FreebaseExtract()

    @timed
    def run(self):
        output = shellout("TMPDIR={tmpdir} ntto -i -a -r {rules} {input} > {output}",
                          tmpdir=tempfile.tempdir, input=self.input().path, rules=self.assets('RULES.txt'))
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='nt'))

class FreebaseJson(FreebaseTask):
    """ Convert all Freebase ntriples to a single JSON file. """

    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return FreebaseExtract()

    @timed
    def run(self):
        output = shellout("TMPDIR={tmpdir} ntto -i -a -j -r {rules} {input} > {output}",
                          tmpdir=tempfile.tempdir, input=self.input().path, rules=self.assets('RULES.txt'))
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'))

class FreebaseEquivalent(FreebaseTask):
    """ Extract all owl:sameAS and equivalent* relations from freebase.

    $ curl -sL rdf.freebase.com/ns/m.0bgd3z7
    @prefix key: <http://rdf.freebase.com/key/>.
    @prefix ns: <http://rdf.freebase.com/ns/>.
    @prefix owl: <http://www.w3.org/2002/07/owl#>.
    @prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#>.
    @prefix xsd: <http://www.w3.org/2001/XMLSchema#>.

    ns:m.0bgd3z7
        ns:base.ontologies.ontology.equivalent_instances    ns:m.0gkkff0;
        ns:base.ontologies.ontology.equivalent_instances    ns:m.0gkkghs;
        ns:base.politeuri.predicate.namespace_1    ns:m.09zs39j;
        key:en    "http_www_w3_org_2002_07_owl_sameas";
        ns:type.object.key    ns:en.http_www_w3_org_2002_07_owl_sameas;
        ns:type.object.name    "http://www.w3.org/2002/07/owl#sameAs"@en;
        ns:type.object.type    ns:base.politeuri.predicate;
        ns:type.object.type    ns:common.topic;
        ns:type.object.type    ns:base.ontologies.topic;
        ns:type.object.type    ns:base.ontologies.ontology;
        ns:rdf:type    ns:common.topic;
        ns:rdf:type    ns:base.ontologies.ontology;
        ns:rdf:type    ns:base.politeuri.predicate;
        ns:rdf:type    ns:base.ontologies.topic;
        rdfs:label    "http://www.w3.org/2002/07/owl#sameAs"@en.

    """
    date = luigi.DateParameter(default=monthly())

    def requires(self):
        return FreebaseExtract(date=self.date)

    @timed
    def run(self):
        output = shellout('LC_ALL="C" LANG=C grep -E "(equivalent|owl#sameAs)" {path} > {output}',
                 path=self.input().path, ignoremap={1: "Not found."})
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='n3'))



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,E1103,C0301
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
Formelly known as directory.mozilla.org.

DMOZ uses a hierarchical ontology scheme for organizing site listings.
Listings on a similar topic are grouped into categories
which can then include smaller categories.
"""

from siskin.benchmark import timed
from gluish.common import Executable
from gluish.intervals import monthly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.task import DefaultTask
import datetime
import luigi

class DMOZTask(DefaultTask):
    """ DMOZ supertask. """
    TAG = 'dmoz'

    def closest(self):
        return monthly(date=self.date)

class DMOZStructureDownload(DMOZTask):
    """ Download structure RDF. """
    date = ClosestDateParameter(default=datetime.date.today())
    url = luigi.Parameter(default='http://rdf.dmoz.org/rdf/structure.rdf.u8.gz', significant=False)

    def requires(self):
        return [Executable(name='curl'), Executable(name='gunzip')]

    @timed
    def run(self):
        output = shellout("""curl "{url}" | gunzip -c > {output}""", url=self.url)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='rdf'))

class DMOZContentDownload(DMOZTask):
    """ Download content RDF. """
    date = ClosestDateParameter(default=datetime.date.today())
    url = luigi.Parameter(default='http://rdf.dmoz.org/rdf/content.rdf.u8.gz', significant=False)

    def requires(self):
        return [Executable(name='curl'), Executable(name='gunzip')]

    @timed
    def run(self):
        output = shellout("""curl "{url}" | gunzip -c > {output}""", url=self.url)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='rdf'))



#!/usr/bin/env python
# coding: utf-8
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
Configuration keys:

[rvk]

url = http://example.org/path/to/rvk.zip
"""

from gluish.format import TSV
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.configuration import Config
from siskin.task import DefaultTask
from siskin.utils import iterfiles
import collections
import datetime
import luigi
import operator
import os
import tempfile

config = Config.instance()

class RVKTask(DefaultTask):
    """ Default RVK task. """
    TAG = 'rvk'

    def closest(self):
        return datetime.date(2014, 11, 1)

class RVKDownload(RVKTask):
    """ Download and unzip XML dump of RVK. """
    url = luigi.Parameter(default=config.get('rvk', 'url', 'http://example.org/path/to/rvk.zip'))
    date = ClosestDateParameter(default=datetime.date.today())

    def run(self):
        target = os.path.join(self.taskdir(), str(self.date))
        if not os.path.exists(target):
            os.makedirs(target)
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        shellout("wget --retry-connrefused -O {stopover} '{url}' && unzip -o -d {dir} {stopover}", dir=target, stopover=stopover, url=self.url)
        files = list(iterfiles(target))
        if not len(files) == 1:
            raise RuntimeError('more than one file')
        luigi.File(files[0]).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='xml', digest=True))

class RVKPaths(RVKTask):
    """ Output path to root, one path per line. """

    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return RVKDownload(date=self.date)

    def run(self):
        output = shellout("xsltproc {xsl} {path} > {output}", xsl=self.assets('rvk.xsl'), path=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path())

class RVKNames(RVKTask):
    """ Output path to root, one path per line. """

    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return RVKDownload(date=self.date)

    def run(self):
        output = shellout("xsltproc {xsl} {path} > {output}", xsl=self.assets('rvktsv.xsl'), path=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class RVKBSZList(RVKTask):
    """ Extract PPN/RVK from current BSZ index. """

    date = luigi.Parameter(default=datetime.date.today())

    def run(self):
        output = shellout("""estab -indices bsz -f "_id content.936.a" > {output} """)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class RVKBSZDistribution(RVKTask):
    """ How many classes are specified. """
    date = luigi.Parameter(default=datetime.date.today())

    def requires(self):
        return RVKBSZList(date=self.date)

    def run(self):
        freq = collections.defaultdict(int)
        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('ppn', 'rvks')):
                for rvk in row.rvks.split('|'):
                    freq[rvk] += 1

        with self.output().open('w') as output:
            for rvk, count in sorted(freq.iteritems(), key=operator.itemgetter(1), reverse=True):
                output.write_tsv(rvk, count)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,E1103,C0301
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
Wikipedia related.
"""

from elasticsearch import helpers as eshelpers
from gluish.common import Executable
from gluish.format import TSV
from gluish.intervals import monthly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from luigi.contrib.esindex import CopyToIndex
from siskin.benchmark import timed
from siskin.task import DefaultTask
from siskin.utils import ElasticsearchMixin
import collections
import datetime
import elasticsearch
import json
import luigi
import urlparse

class WikipediaTask(DefaultTask):
    """
    Generic wikipedia related task.
    """
    TAG = 'wikipedia'

    def closest(self):
        return monthly(self.date)

class WikipediaArticleDump(WikipediaTask):
    """
    Download wiki-latest-pages-articles.xml.bz2 for a given language. Unarchive on the fly.
    """
    language = luigi.Parameter(default='en')
    date = ClosestDateParameter(default=datetime.date.today())

    @timed
    def run(self):
        host = 'http://dumps.wikimedia.org'
        path = '/{lc}wiki/latest/{lc}wiki-latest-pages-articles.xml.bz2'.format(
                lc=self.language)
        url = urlparse.urljoin(host, path)
        output = shellout("curl {url} | bunzip2 -c > {output}", url=url)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='xml'))

class WikipediaOutlinks(WikipediaTask):
    """ Collect links to the interwebs (best effort). """
    language = luigi.Parameter(default='en')
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return WikipediaArticleDump(language=self.language, date=self.date)

    @timed
    def run(self):
        """
        Wrong Assumption: There is no space or | in the URL.
        """
        output = shellout("""LANG=C grep -o -E "http://[^| ]*" {input} > {output}""", input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class WikipediaTitles(WikipediaTask):
    """
    Get a list of wikipedia article titles.

    James Joyce
    Judo
    JapaneseLanguage
    James Bond
    Japanese language
    Johnny Got His Gun
    """
    language = luigi.Parameter(default='en')
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return WikipediaArticleDump(date=self.date, language=self.language)

    @timed
    def run(self):
        output = shellout("""LANG=C grep -E "<title>[^<]*</title>" {input} |
                             LANG=C sed -e 's@<title>@@g' |
                             LANG=C sed -e 's@</title>@@g' | sed -e 's/^ *//g' > {output}""",
                             input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='tsv'), format=TSV)

class WikipediaTitlesInBSZ(WikipediaTask):
    """ Run a query with the exact wikipedia title in elasticsearch/bsz
    and report back the number of hits. """

    language = luigi.Parameter(default='en')
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return WikipediaTitles(date=self.date, language=self.language)

    @timed
    def run(self):
        es = elasticsearch.Elasticsearch()
        with self.input().open() as handle:
            with self.output().open('w') as output:
                for row in handle.iter_tsv(cols=('title',)):
                    title = row.title.strip()
                    try:
                        result = es.search(index='bsz', q='"%s"' % title.replace('"','\\"'))
                        output.write_tsv(title, result['hits']['total'])
                    except Exception as exc:
                        self.logger.warn(exc)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='tsv'), format=TSV)

class WikipediaIsbnList(WikipediaTask):
    """
    A list ISBN (candidates) found in the dump. Contains dups.

    9780716703440
    9780471925675
    9780486653839
    9780486653839
    9780674033634
    9780201151428
    9780805386622
    9780521316774
    9780201149425
    9780767915014
    """

    language = luigi.Parameter(default='en')
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {'dump': WikipediaArticleDump(date=self.date, language=self.language),
                'app': Executable(name='isbngrep')}

    @timed
    def run(self):
        output = shellout("cat {input} | isbngrep -n > {output}",
                          input=self.input().get('dump').path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='tsv'), format=TSV)

class WikipediaCategoryList(WikipediaTask):
    """
    A list of wikipedia categories. Using `grep`.

    ...
    Category:1358 establishments by country
    Category:1358 establishments in England
    Category:1358 in England
    Category:1358 in Europe
    Category:1358 in France
    Category:1358 in Norway
    Category:1358 in international relations
    Category:1358 in law
    Category:1358 in politics
    Category:1358 works
    ...
    """

    language = luigi.Parameter(default='en')
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return WikipediaTitles(date=self.date, language=self.language)

    @timed
    def run(self):
        prefixes = {'en': 'Category', 'de': 'Kategorie', 'fr': u'Catégorie'}
        if self.language not in prefixes:
            raise RuntimeError('Categorie prefix not added yet')
        output = shellout("""LANG=C grep -iF "{prefix}:" {input} > {output}""",
                          input=self.input().path,
                          prefix=prefixes.get(self.language))
        output = shellout("LANG=C sort {input} | LANG=C uniq > {output}", input=output)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='tsv'), format=TSV)

class WikipediaJson(WikipediaTask):
    """
    A Json representation of a wikipedia dump. Line delimited JSON.
    """

    language = luigi.Parameter(default='en')
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {'dump': WikipediaArticleDump(date=self.date, language=self.language),
                'app': Executable(name='wikitojson', message='https://github.com/miku/wikitools')}

    @timed
    def run(self):
        output = shellout("wikitojson {input} > {output}", input=self.input().get('dump').path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'))

class WikipediaCategoryTable(WikipediaTask):
    """
    A list of wikipedia categories. Using wikitools
    (https://github.com/miku/wikitools).

    Similar to `WikipediaCategoryList`.

    Example output:

        ...
        Elia Kazan  Autobiografie
        Elia Kazan  Drehbuchautor
        Elia Kazan  Oscarpreisträger
        ...

    Approximate sizes:

        English: 21759052
        German:   6927106
        French:   5682380
        Spanish:  3365340
    """

    language = luigi.Parameter(default='en')
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {'dump': WikipediaArticleDump(date=self.date, language=self.language),
                'app': Executable(name='wikicats', message='https://github.com/miku/wikitools')}

    @timed
    def run(self):
        prefixes = {
            'de': 'Kategorie',
            'en': 'Category',
            'fr': u'Catégorie',
            'eo': 'Kategorio',
            'es': u'Categoría',
            'it': 'Categoria',
            'he': u'קטגוריות', # won't work
        }
        if self.language not in prefixes:
            raise RuntimeError('Category prefix not added yet')

        output = shellout("""wikicats -pattern "{prefix}" {input} > {output}""",
                          input=self.input().get('dump').path, prefix=prefixes.get(self.language),
                          encoding='utf-8')
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class WikipediaCategoryExtension(WikipediaTask):
    """
    For each category, collect the pages it contains. Similar to
    WikipediaCategoryExtensionJson, except the output here is tabular.

    Badeanlage in Niedersachsen Moskaubad|Maschsee|Hardausee|Waldbad Birkerteich
    Filmpreis als Thema Kategorie:Goldene Kamera|Kategorie:Golden Globe Award
    Kanute (Schweiz)    Daniela Baumer|Corrado Filipponi|Ron Fischer (Kanute)|Michael Kurt
    """

    language = luigi.Parameter(default='de')
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return WikipediaCategoryTable(date=self.date, language=self.language)

    @timed
    def run(self):
        reverse = collections.defaultdict(set)
        with self.input().open() as handle:
            for row in handle.iter_tsv():
                if not len(row) == 2:
                    continue
                page, category = row
                reverse[category].add(page)
        with self.output().open('w') as output:
            for category, pages in reverse.iteritems():
                output.write_tsv(category, '|'.join(pages))

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class WikipediaCategoryExtensionJson(WikipediaTask):
    """
    For each category, collect the pages it contains.
    Example:

    {
       "category" : "Jazz-Oboist",
       "pages" : [
          "Bob Cooper",
          "Charles Owens (Saxophonist, 1939)",
          "Karl Jenkins",
          "Marshall Allen",
          "Matthias Schubert",
          "Maud Sauer",
          "Paul McCandless",
          "Phil Bodner",
          "Roger Janotta",
          "Sonny Simmons",
          "Werner Baumgart",
          "Yusef Lateef"
       ]
    }
    """

    language = luigi.Parameter(default='de')
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return WikipediaCategoryTable(date=self.date, language=self.language)

    @timed
    def run(self):
        reverse = collections.defaultdict(set)
        with self.input().open() as handle:
            for row in handle.iter_tsv():
                if not len(row) == 2:
                    continue
                page, category = row
                reverse[category].add(page)
        with self.output().open('w') as output:
            for category, pages in reverse.iteritems():
                output.write(json.dumps(dict(category=category, pages=sorted(pages))))
                output.write('\n')

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'))

class WikipediaCategoryExtensionIndex(WikipediaTask, CopyToIndex):
    """
    Index the page title authority id docs.
    """

    language = luigi.Parameter(default='de')
    date = ClosestDateParameter(default=datetime.date.today())

    @property
    def index(self):
        return '%swikicats' % self.language

    doc_type = 'default'

    mapping = {'default': {'date_detection': False,
                           '_all': {'enabled': True,
                                    'term_vector': 'with_positions_offsets',
                                    'store': True}}}

    def requires(self):
        return WikipediaCategoryExtensionJson(language=self.language, date=self.date)

class WikipediaCategoryDistribution(WikipediaTask):
    """
    Something like this: https://gist.github.com/miku/78d84756d08d7dace204

    Another example from the english wikipedia:

         664354 Living people
          88726 Archived files for deletion discussions
          51015 Year of birth missing (living people)
          37800 Pending DYK nominations
          28772 English-language films
          20531 American films
          19743 Local COIBot Reports
          19606 Year of birth unknown
          18093 Year of birth missing
          17793 The Football League players
          17192 Year of death missing
        ...
    """
    language = luigi.Parameter(default='en')
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return WikipediaCategoryTable(date=self.date, language=self.language)

    @timed
    def run(self):
        output = shellout(r"""cat {input} | cut -f2- | LANG=C sort | LANG=C uniq -c |
                              LANG=C sort -nr > {output}""", input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class WikipediaPagesWithCategories(WikipediaTask):
    """
    Contains dirty rows. But good row look like this:

    ...
    1932 Maccabiah Games
    1932 Madras and Southern Mahratta Railway Strike
    1932 Major League Baseball season
    1932 Massachusetts State Aggies football team
    1932 Memorial Cup
    ...
    A Gleam
    A Gleam Handicap
    A Glen Campbell Christmas
    A Glimmer of Hope
    ...
    Abazu-Akabo
    Abaí
    Abaíra
    Abaúj (region)
    ...
    """

    language = luigi.Parameter(default='en')
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return WikipediaCategoryTable(date=self.date, language=self.language)

    @timed
    def run(self):
        output = shellout(r"""cat {input} | cut -f1 | LANG=C sort | LANG=C uniq > {output}""",
                          input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class WikipediaIndex(WikipediaTask, CopyToIndex):
    language = luigi.Parameter(default='en')
    date = ClosestDateParameter(default=datetime.date.today())

    purge_existing_index = True

    @property
    def index(self):
        return '%swiki' % self.language

    doc_type = 'page'

    def requires(self):
        return WikipediaJson(language=self.language, date=self.date)

#
# Only for the German (Authority data)
#

class WikipediaRawAuthorityData(WikipediaTask):
    language = luigi.Parameter(default='de')
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {'dump': WikipediaArticleDump(date=self.date, language=self.language),
                'app': Executable(name='wikinorm', message='https://github.com/miku/wikitools')}

    @timed
    def run(self):
        prefixes = {'de': 'Normdaten', 'en': 'Authority control'}
        if not self.language in prefixes:
            raise RuntimeError("Languages available %s" % prefixes)
        output = shellout("""wikinorm -pattern '{prefix}' {input} > {output}""",
                          input=self.input().get('dump').path, prefix=prefixes.get(self.language))
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class WikipediaRawAuthorityJson(WikipediaTask):
    language = luigi.Parameter(default='de')
    date = ClosestDateParameter(default=datetime.date.today())
    allowed_errors = luigi.IntParameter(default=100, significant=False)

    def requires(self):
        return WikipediaRawAuthorityData(date=self.date, language=self.language)

    @timed
    def run(self):
        errors = 0
        with self.input().open() as handle:
            with self.output().open('w') as output:
                for row in handle.iter_tsv():
                    if not len(row) == 2:
                        errors += 1
                        if errors > self.allowed_errors:
                            raise RuntimeError('more than %s errors, adjust allowed_errors')
                        else:
                            self.logger.error('error in row: {0}'.format(str(row)))
                            continue
                    title, ids = row
                    trimmed = ids.strip("{}")
                    authority = {'title': title}
                    for entry in trimmed.split('|'):
                        parts = entry.split("=")
                        if not len(parts) == 2:
                            continue
                        name, value = parts
                        authority[name.lower()] = value
                    output.write(json.dumps(authority))
                    output.write("\n")

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'))

class WikipediaGNDList(WikipediaTask):
    """
    TSV(title, GND)
    """

    language = luigi.Parameter(default='de')
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return WikipediaRawAuthorityJson(date=self.date, language=self.language)

    @timed
    def run(self):
        with self.input().open() as handle:
            with self.output().open('w') as output:
                for line in handle:
                    content = json.loads(line)
                    if 'gnd' in content:
                        output.write_tsv(content.get('title', 'NA').encode('utf-8'),
                                         content.get('gnd', 'NA').encode('utf-8'))

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class WikipediaRawAuthorityIndex(WikipediaTask, CopyToIndex):
    """
    Index the page title authority id docs.
    """

    language = luigi.Parameter(default='de')
    date = ClosestDateParameter(default=datetime.date.today())

    index = 'dewikinorm'
    doc_type = 'default'

    mapping = {'default': {'date_detection': False,
                           '_all': {'enabled': True,
                                    'term_vector': 'with_positions_offsets',
                                    'store': True}}}

    def requires(self):
        return WikipediaRawAuthorityJson(language=self.language, date=self.date)

class WikipediaCategoryGNDList(WikipediaTask, ElasticsearchMixin):
    """
    Important. For each category, find the associated pages and find the GND
    lookup the GND of that page.
    """

    language = luigi.Parameter(default='de')
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {'norm': WikipediaRawAuthorityIndex(date=self.date,
                                                   language=self.language),
                'cats': WikipediaCategoryExtensionIndex(date=self.date,
                                                        language=self.language)}
    def run(self):
        es = elasticsearch.Elasticsearch(timeout=30)
        hits = eshelpers.scan(es, {'query': {'match_all': {}},
            'fields': ["category", "pages"]}, index='dewikicats', doc_type='default',
            scroll='20m', size=100000)
        with self.output().open('w') as output:
            for hit in hits:
                fields = hit.get('fields')
                output.write_tsv(hit['_id'], fields['category'][0].encode('utf-8'))

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class WikipediaCategoryExample(WikipediaTask, ElasticsearchMixin):
    """
    See whats in Komponisten der Romantik: NqFMwG6rTYK6yqVpQGF_tQ
    """
    category = luigi.Parameter(default='NqFMwG6rTYK6yqVpQGF_tQ')

    def run(self):
        gnds, empty, i = set(), 0, 0
        es = elasticsearch.Elasticsearch()
        result = es.get(index='dewikicats', id=self.category)

        for i, page in enumerate(result.get('_source').get('pages')):
            self.logger.info(page)
            r = es.search(index='dewikinorm', body={'query': {'query_string': {'query': '"%s"' % page}}})
            for hit in r['hits']['hits']:
                self.logger.info(hit.get('_source').get('gnd'))
                if hit.get('_source').get('gnd'):
                    gnds.add(hit.get('_source').get('gnd'))
                else:
                    empty += 1

        self.logger.info('{0} {1} {2}'.format(len(gnds), empty, i))

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'))

class WikipediaRawCitations(WikipediaTask):
    """
    Very crudely extract citations from wikipedia markup.

    Example:

    {{cite book | url=http://books.google.de/books?id=GlICpnVArb8C&amp;pg=PA304 | pages =304–305 | chapter =The Antimonials ...
    {{cite book|author = [[Michael Sipser]] | year = 1997 | title = Introduction to the Theory of Computation
    ...
    """
    language = luigi.Parameter(default='en')
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return WikipediaArticleDump(date=self.date, language=self.language)

    @timed
    def run(self):
        output = shellout("""LANG=C egrep -oi "\\{{\\{{cite book[^}}]*?}}}}" {input} > {output}""", input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='txt'))



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,E1103,C0301
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
EBL.

Configuration keys:

[ebl]

ftp-host = host.name
ftp-username = username
ftp-password = password
ftp-path = /
ftp-pattern = some*glob*pattern.zip

# Dates to ignore.
exclude = 1970-12-31

# This regexes are matched against the file paths. It must provide `year`, `month` and `day` named groups.
regex-dump = .*(?P<year>\d{4})(?P<month>\d{2})(?P<day>\d{2})-(\d{4})_leip_FULL_CATALOGUE_export.zip
regex-delta = .*(?P<year>\d{4})(?P<month>\d{2})(?P<day>\d{2})-(\d{4})_leip_Content_export.zip

glob-add = *leip_Content_Add_*.mrc
glob-delete = *leip_Content_Delete_*.mrc
"""

from gluish.format import TSV
from gluish.intervals import hourly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from luigi.contrib.esindex import CopyToIndex
from siskin.benchmark import timed
from siskin.common import FTPMirror, Executable
from siskin.configuration import Config
from siskin.database import sqlitedb
from siskin.task import DefaultTask
from siskin.utils import memoize
import base64
import datetime
import luigi
import operator
import re
import tempfile

config = Config.instance()

class EBLTask(DefaultTask):
    TAG = '004'

    @memoize
    def muted(self):
        """ Parse ebl.config and return a set of dates to ignore. """
        try:
            dates = config.get('ebl', 'exclude', '').split(',')
            return set([datetime.date(*(int(v) for v in s.split('-'))) for s in dates])
        except ValueError as err:
            raise RuntimeError('could not parse ebl.exclude: %s' % err)

    @memoize
    def closest(self):
        """ Just use the output of `EBLLatestDate` here.
        Any exception occuring in `EBLLatestDate` will propagate,
        e.g. if a too early date is requested. """
        if not hasattr(self, 'date'):
            raise RuntimeError('cannot find closest date for a task w/o date')
        task = EBLLatestDate(date=self.date)
        luigi.build([task], local_scheduler=True)
        s = task.output().open().read().strip()
        result = datetime.date(*(int(v) for v in s.split('-')))
        return result

class EBLPaths(EBLTask):
    """ A list of EBL file paths (via FTP). """
    indicator = luigi.Parameter(default=hourly(fmt='%s'))

    def requires(self):
        host = config.get('ebl', 'ftp-host')
        username = config.get('ebl', 'ftp-username')
        password = config.get('ebl', 'ftp-password')
        base = config.get('ebl', 'ftp-path')
        pattern = config.get('ebl', 'ftp-pattern')
        return FTPMirror(host=host, username=username, password=password,
                         base=base, pattern=pattern)

    @timed
    def run(self):
        self.input().move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class EBLInventory(EBLTask):
    """ List EBL inventory in form of (type, date, path) tuples. No sorting. """
    indicator = luigi.Parameter(default=hourly(fmt='%s'))

    def requires(self):
        return EBLPaths()

    def run(self):
        getdate = operator.itemgetter('year', 'month', 'day')
        patterns = (
            ('dump', re.compile(config.get('ebl', 'regex-dump'))),
            ('delta', re.compile(config.get('ebl', 'regex-delta'))),
        )
        with self.input().open() as handle:
            with self.output().open('w') as output:
                for row in handle.iter_tsv(cols=('path',)):
                    for name, pattern in patterns:
                        match = pattern.search(row.path)
                        if match:
                            date = datetime.date(*map(int, getdate(match.groupdict())))
                            if date in self.muted():
                                continue
                            output.write_tsv(name, date, row.path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class EBLBacklog(EBLTask):
    """ A list of all task relevant for a certain date. Includes dumps and deltas.
    Will raise an error, if there no dump can be found. Ordered by date, dump first. """

    date = luigi.DateParameter(default=datetime.date.today())

    def requires(self):
        return EBLInventory()

    def run(self):
        backlog = []
        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('kind', 'date', 'path')):
                date = datetime.date(*map(int, row.date.split('-')))
                if date <= self.date:
                    backlog.append(row)
        if all([entry.kind == 'delta' for entry in backlog]):
            raise RuntimeError('No EBL dump found before %s' % self.date)
        backlog = sorted(backlog, key=operator.itemgetter(1), reverse=True)
        index = [entry.kind for entry in backlog].index('dump') + 1
        backlog = reversed(backlog[:index])
        with self.output().open('w') as output:
            for entry in backlog:
                output.write_tsv(*entry)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class EBLDumpCombined(EBLTask):
    """ Combine .mrc files out of dump. """
    date = luigi.DateParameter(default=datetime.date.today())

    def requires(self):
        return EBLInventory()

    def run(self):
        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('kind', 'date', 'path')):
                date = datetime.date(*map(int, row.date.split('-')))
                if date == self.date and row.kind == 'dump':
                    output = shellout("unzip -p {input} \*.mrc > {output}", input=row.path,
                                      ignoremap={1: 'A warning alone will trigger a non-zero return value. Assuming everything went well anyway.'})
                    luigi.File(output).move(self.output().path)
                    break
            else:
                raise RuntimeError("No EBL dump on given date: %s" % self.date)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='mrc'))

class EBLDeltaCombined(EBLTask):
    """ Out of a delta, extract and combine the additions or deletions. """

    date = luigi.DateParameter(default=datetime.date.today())
    kind = luigi.Parameter(default='add', description='add or delete')

    def requires(self):
        return EBLInventory()

    def run(self):
        patterns = {
            'add': config.get('ebl', 'glob-add'),
            'delete': config.get('ebl', 'glob-delete'),
        }
        if self.kind not in patterns:
            raise RuntimeError("Unknown kind, only valid kinds are add and delete.")
        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('kind', 'date', 'path')):
                date = datetime.date(*map(int, row.date.split('-')))
                if date == self.date and row.kind == 'delta':
                    output = shellout("unzip -p {input} \{pattern} > {output}", input=row.path, pattern=patterns[self.kind],
                                      ignoremap={1: 'A warning alone will trigger a non-zero return value. Assuming everything went well anyway.'})
                    luigi.File(output).move(self.output().path)
                    break
            else:
                raise RuntimeError("No EBL delta package on given date: %s" % self.date)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='mrc'))

class EBLMarcDB(EBLTask):
    """ Create a sqlite3 db with (id, secondary (date), blob) table to allow random access
    to a single marc records by id for a certain date. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return EBLBacklog(date=self.closest())

    @timed
    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('kind', 'date', 'path')):
                date = datetime.date(*map(int, row.date.split('-')))
                if row.kind == 'delta':
                    task = EBLDeltaCombined(date=date, kind='add')
                    luigi.build([task])
                    shellout("marcdb -encode -secondary {date} -o {output} {input}", date=date, input=task.output().path, output=stopover)
                if row.kind == 'dump':
                    task = EBLDumpCombined(date=date)
                    luigi.build([task])
                    shellout("marcdb -encode -secondary {date} -o {output} {input}", date=date, input=task.output().path, output=stopover)
        luigi.File(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='db'))

class EBLEvents(EBLTask):
    """ Given a date, create a single MARC file, that incorporates the dump and all deltas (additions, deletions). """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return EBLBacklog(date=self.closest())

    @timed
    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('kind', 'date', 'path')):
                date = datetime.date(*map(int, row.date.split('-')))
                if row.kind == 'dump':
                    task = EBLDumpCombined(date=date)
                    luigi.build([task])
                    shellout("marctotsv {input} U 001 {date} >> {output}", input=task.output().path, date=date, output=stopover)
                if row.kind == 'delta':
                    task = EBLDeltaCombined(date=date, kind='add')
                    luigi.build([task])
                    shellout("marctotsv {input} U 001 {date} >> {output}", input=task.output().path, date=date, output=stopover)
                    task = EBLDeltaCombined(date=date, kind='delete')
                    luigi.build([task])
                    shellout("marctotsv {input} D 001 {date} >> {output}", input=task.output().path, date=date, output=stopover)
        luigi.File(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class EBLSurface(EBLTask):
    """ For a given date, list the (EBL ID, date) that should be in the current
    state of the data source. """

    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return EBLEvents(date=self.date)

    @timed
    def run(self):
        datemap = {}
        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('kind', 'id', 'date')):
                date = datetime.date(*map(int, row.date.split('-')))
                if row.kind == 'U':
                    datemap[row.id] = date
                if row.kind == 'D':
                    if row.id in datemap:
                        del datemap[row.id]
                    else:
                        # just another indication of a deleted record that
                        # was not added before - ignore
                        pass

        with self.output().open('w') as output:
            for id, date in sorted(datemap.iteritems(), key=operator.itemgetter(1)):
                output.write_tsv(id, date)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class EBLSnapshot(EBLTask):
    """ Create a single MARC file, that represents the state of the data for
    a given date. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {'surface': EBLSurface(date=self.date),
                'db': EBLMarcDB(date=self.date)}

    @timed
    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        with sqlitedb(self.input().get('db').path) as cursor:
            with self.input().get('surface').open() as handle:
                with open(stopover, 'wb') as output:
                    for row in handle.iter_tsv(cols=('id', 'date')):
                        cursor.execute("SELECT record from store where id = ? and secondary = ?", (row.id, row.date))
                        result = cursor.fetchone()
                        output.write(base64.b64decode(result[0]))
        luigi.File(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='mrc'))

class EBLLatestDateAndPath(EBLTask):
    """ For a given date, return the closest date in the past
    on which EBL shipped (dump or delta). """
    date = luigi.DateParameter(default=datetime.date.today())
    indicator = luigi.Parameter(default=hourly(fmt='%s'))

    def requires(self):
        return EBLInventory(indicator=self.indicator)

    @timed
    def run(self):
        """ load dates and paths in pathmap, use a custom key
        function to find the closest date, but double check,
        if that date actually lies in the future - if it does, raise
        an exception, otherwise dump the closest date and filename to output """
        pathmap = {}
        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('kind', 'date', 'path')):
                date = datetime.datetime.strptime(row.date, '%Y-%m-%d').date()
                pathmap[date] = row.path

        def closest_keyfun(date):
            if date > self.date:
                return datetime.timedelta(1e8)
            return abs(date - self.date)

        closest = min(pathmap.keys(), key=closest_keyfun)
        if closest > self.date:
            raise RuntimeError('No shipment before: %s' % min(pathmap.keys()))

        with self.output().open('w') as output:
            output.write_tsv(closest, pathmap.get(closest))

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class EBLLatestDate(EBLTask):
    """ Only output the latest date. """
    date = luigi.DateParameter(default=datetime.date.today())
    indicator = luigi.Parameter(default=hourly(fmt='%s'))

    def requires(self):
        return EBLLatestDateAndPath(date=self.date, indicator=self.indicator)

    @timed
    def run(self):
        output = shellout("awk '{{print $1}}' {input} > {output}",
                          input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class EBLJson(EBLTask):
    """ Take EBL combined and convert it to JSON. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {'marc': EBLSnapshot(date=self.date),
                'converter': Executable(name='marctojson', message='http://git.io/1LXpQA')}

    @timed
    def run(self):
        output = shellout("marctojson -m date={date} {input} > {output}",
                          input=self.input().get('marc').fn,
                          date=self.closest())
        luigi.File(output).move(self.output().fn)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='json'))

class EBLIndex(EBLTask, CopyToIndex):
    """ Index EBL. """
    date = ClosestDateParameter(default=datetime.date.today())

    index = 'ebl'
    doc_type = 'title'
    purge_existing_index = True

    settings = {
        "settings": {
            "number_of_shards": 5,
            "number_of_replicas": 0,
            "analysis": {
                "filter": {
                    "autocomplete_filter": {
                        "type":     "edge_ngram",
                        "min_gram": 1,
                        "max_gram": 20
                    }
                },
                "analyzer": {
                    "autocomplete": {
                        "type":      "custom",
                        "tokenizer": "standard",
                        "filter": [
                            "lowercase",
                            "autocomplete_filter"
                        ]
                    }
                }
            }
        }
    }

    mapping = {
        'title': {
            'date_detection': False,
            '_id': {
                'path': 'content.001'
            },
            '_all': {
                'enabled': True,
                'term_vector': 'with_positions_offsets',
                'store': True
            },
            'properties': {
                'content': {
                    'properties': {
                        '245': {
                            'properties': {
                                'a': {
                                    'type': 'string',
                                    'index_analyzer': 'autocomplete',
                                    'search_analyzer': 'standard'
                                }
                            }
                        },
                        '100': {
                            'properties': {
                                'a': {
                                    'type': 'string',
                                    'index_analyzer': 'autocomplete',
                                    'search_analyzer': 'standard'
                                }
                            }
                        },
                        '700': {
                            'properties': {
                                'a': {
                                    'type': 'string',
                                    'index_analyzer': 'autocomplete',
                                    'search_analyzer': 'standard'
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    def update_id(self):
        """ This id will be a unique identifier for this indexing task."""
        return self.effective_task_id()

    def requires(self):
        return EBLJson(date=self.date)



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,E1103
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
Gutenberg metadata.
"""

from luigi.contrib.esindex import CopyToIndex
from gluish.intervals import weekly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.task import DefaultTask
import datetime
import luigi

class GutenbergTask(DefaultTask):
    TAG = '001'

    def closest(self):
        return weekly(date=self.date)

class GutenbergDownload(GutenbergTask):
    """ Download and extract on the fly. """
    date = ClosestDateParameter(default=datetime.date.today())
    url = luigi.Parameter(default="http://www.gutenberg.org/feeds/catalog.marc.bz2", significant=False)

    def run(self):
        output = shellout("curl -L {url} | bunzip2 -c > {output}", url=self.url)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='mrc'))

class GutenbergJson(GutenbergTask):
    """ Convert Marc to Json. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return GutenbergDownload(date=self.date)

    def run(self):
        output = shellout("marctojson -m date={date} {input} > {output}",
                          date=self.closest(), input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'))

class GutenbergIndex(GutenbergTask, CopyToIndex):
    date = ClosestDateParameter(default=datetime.date.today())

    index = 'gutenberg'
    doc_type = 'title'
    purge_existing_index = True

    mapping = {'title': {'date_detection': False,
                          '_id': {'path': 'content.001'},
                          '_all': {'enabled': True,
                                   'term_vector': 'with_positions_offsets',
                                   'store': True}}}

    def update_id(self):
        """ This id will be a unique identifier for this indexing task."""
        return self.effective_task_id()
    def requires(self):
        return GutenbergJson(date=self.date)



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,E1103
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

from gluish.intervals import monthly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.benchmark import timed
from siskin.task import DefaultTask
import datetime
import luigi
import tempfile

class HathiTask(DefaultTask):
    TAG = '035'

    def closest(self):
        return monthly(date=self.date)


class HathiDump(HathiTask):
    """ Dump SSOAR. """

    def requires(self):
        return Executable(name='oaimi', message='https://github.com/miku/oaimi')

    def run(self):
        output = shellout("""
            cat <(echo '<collection xmlns="http://www.openarchives.org/OAI/2.0/" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">')
                <(oaimi -verbose -prefix marc21 -set hathitrust http://quod.lib.umich.edu/cgi/o/oai/oai)
                <(echo '</collection>') > {output}""")
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path())

class HathiCombine(HathiTask):
    """ Combine the chunks for a date range into a single file.
    The filename will carry a name, that will only include year
    and month for begin and end. """

    def requires(self):
        return SSOARDump()

    @timed
    def run(self):
        output = shellout("""yaz-marcdump -f utf-8 -t utf-8 -i marcxml -o marc {input} > {output}""", input=self.input().path, ignoremap={5: 'TODO: fix this'})
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='mrc'))

class HathiJson(HathiTask):
    """ Json version. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return SSOARCombine(date=self.date)

    def run(self):
        output = shellout("marctojson -m date={date} {input} > {output}",
                          input=self.input().path, date=self.closest())
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'))